#include <stdio.h>
#include <stdint.h>
#include <gmp.h>
#include "cuda_helpers.h"

#define DIM (1<<20) 
#define NPRIMES 5

#define NLIMBS_LOW 11
#define NLIMBS 16
#define LS 65536


/* Multiply
 * (c1,c0) = a*b            */
#define mul(c1,c0,a,b)                             \
do {                                               \
  c0 = a*b;                                        \
  c1 = __umulhi(a,b);                              \
} while (0)

// TODO: Can be improved with inline asm even further
/* Multiply-and-accumulate
 * (c1,c0) = a*b+c0         */
#define muladd(c1,c0,a,b)                          \
do {                                               \
  uint32_t __ma = c0;                              \
  mul(c1,c0,a,b);                                  \
  c0 = c0 + __ma;                                  \
  c1 = c1 + (c0 < __ma);                           \
} while (0)


__device__ void mult(uint32_t *b, uint32_t a) {
  uint32_t c0, c1, c2;
  
  mul (c1,c0,a,b[0]);
  muladd(c2,c1,a,b[1]);
  b[0] = c0;
  b[1] = c1;

  muladd(c0,c2,a,b[2]);
  muladd(c1,c0,a,b[3]);
  b[2] = c2;
  b[3] = c0;

  muladd(c2,c1,a,b[4]);
  muladd(c0,c2,a,b[5]);
  b[4] = c1;
  b[5] = c2;
  
  muladd(c1,c0,a,b[6]);
  muladd(c2,c1,a,b[7]);
  b[6] = c0;
  b[7] = c1;
  
  muladd(c0,c2,a,b[8]);
  muladd(c1,c0,a,b[9]);
  b[8] = c2;
  b[9] = c0;
  
  muladd(c2,c1,a,b[10]);
  muladd(c0,c2,a,b[11]);
  b[10] = c1;
  b[11] = c2;

  muladd(c1,c0,a,b[12]);
  muladd(c2,c1,a,b[13]);
  b[12] = c0;
  b[13] = c1;
  
  muladd(c0,c2,a,b[14]);
  b[14] = c2;
  b[15] = c0;
}

__device__ void zero(unsigned int * num) {
  for(int i=0; i<NLIMBS; i++) {
    num[i] = 0;
  }
}

__device__ void copy(unsigned int * out, unsigned int * in) {
  for(int i=0; i<NLIMBS; i++) {
    out[i] = in[i];
  }
}

#define too_long(num) ((num)[NLIMBS-1] != 0)

// TODO: This function should be optimized 
// --> turn into straight-line code
__device__ bool is_good(unsigned int * num) {
  int i, v;
  
  v = num[11]|num[12]|num[13]|num[14]|num[15];
  // When v = 0 return false --> too small.

  if (v == 0) {
    return false;
  }
  
  for (i=15; i >= 11; i--) {
    if (num[i] != 0) break;
  }

  for(; i >= 0; i--) {
    if (num[i] == 0) {
      return true;
    }
  }
  return false;
}

__device__ void check_result(unsigned int * num, unsigned int * outbuf, unsigned int * outcnt) {
  if(is_good(num)) {
    unsigned int old_outcnt = atomicAdd(outcnt, 1);
    if (*outcnt < DIM) {
      copy(&outbuf[old_outcnt * NLIMBS], num);
    }
  }
}

__device__ void dostuff(unsigned int* outbuf, unsigned int* outcnt, uint32_t *list, unsigned int* primes, unsigned int start) {
  // initialize all prime factors 1, ie the exponents to 0
  uint32_t cur[NPRIMES][NLIMBS], *lcur;
  uint32_t l0 = list[2*start];
  uint32_t l1 = list[2*start+1];

  //if (list[2*start] == 0) {  
  //  unsigned int old_outcnt = atomicAdd(outcnt, 1);
  //  outbuf[old_outcnt * NLIMBS] = start; 
  //  outbuf[old_outcnt * NLIMBS+1] = (list[0] == 0);
  //  return;
  //} 

  for (int j=0; j < NPRIMES; j++) {
    cur[j][0] = l0; 
    cur[j][1] = l1;
    for (int i=2; i < NLIMBS; i++) {
      cur[j][i] = 0;
    }
  }

  // initialize cur[0] to prime^start
  //for (int j=0; j < start; j++) {
  //  mult (cur[0], primes[0]);
  //}

  // initialize cur[j] to the value of cur[0]
  //for (int j=1; j < NPRIMES; j++) {
  //  copy (cur[j], cur[0]);
  //}

  lcur = cur[NPRIMES-1];

  while (1) {
    check_result(lcur, outbuf, outcnt);

    // increase number at index by the corresponding prime
    mult (lcur, primes[NPRIMES-1]);

    // exceeded upper bound, same check as above but now with one more
    if (too_long(lcur)) {
      int index = NPRIMES-1;
      do {

        // we have increased all numbers to the maximum => done
        if (index == 0) {
          return;
        }
        
	// go to next index and increase it
        index--;

        // increase current index by the appropriate prime
        mult (cur[index], primes[index]);
      } while (too_long(cur[index]));

      // reset all lower positions to current highest prime power
      // effectively setting their power to 0
      for (int j=index+1; j < NPRIMES; j++) {
        copy (cur[j], cur[index]);
      }
    }
  }
}


__global__ void add(unsigned int *outbuf, unsigned int *outcnt, uint32_t *list, unsigned int *primes) {
    //dostuff(outbuf, outcnt, primes, threadIdx.x);
    dostuff(outbuf, outcnt, list, primes, threadIdx.x + blockDim.x * blockIdx.x);
}


int main(void) {
  unsigned int *outbuf;
  unsigned int outcnt = 0;
  unsigned int primes[NPRIMES] = {13, 11, 7, 5, 3};
    //unsigned int primes[NPRIMES] = {7, 5, 3};

  outbuf = (uint32_t*) malloc(DIM*NLIMBS*sizeof(uint32_t));

  unsigned int *dev_outbuf = (unsigned int *)CUDA_MALLOC(DIM * NLIMBS * sizeof(unsigned int));
  unsigned int *dev_outcnt = (unsigned int *)CUDA_MALLOC(sizeof(unsigned int));
  unsigned int *dev_primes = (unsigned int *)CUDA_MALLOC(NPRIMES * sizeof(unsigned int));

#define BLOCKS 32
#define TOTAL 256
    
  uint32_t *dev_list = (uint32_t*)CUDA_MALLOC(BLOCKS * TOTAL * 2 * sizeof(uint32_t));

  uint32_t *enum_list; //[TOTAL];
  enum_list = (uint32_t*) malloc (TOTAL*BLOCKS*2*sizeof(uint32_t));

{
  FILE *fp;
  char *line = NULL;
  size_t len = 0;
  mpz_t num;
  mpz_init (num);
  int i=0;


  fp = fopen("genlist.txt", "r");
  if (fp == NULL)
    exit(EXIT_FAILURE);

  while ((getline(&line, &len, fp)) != -1 && i < BLOCKS*TOTAL) {
    mpz_set_str (num, line, 10);
    enum_list[2*i] = mpz_get_ui (num);
    mpz_fdiv_q_2exp (num, num, 32);
    enum_list[2*i+1] = mpz_get_ui (num); 
    printf ("%x %x\n", enum_list[2*i+1], enum_list[2*i]);
    i++;
  }

  free(line);
  mpz_clear (num);
  fclose(fp);
  
  system ("cp genlist.txt genlist.txt.backup");
  system ("sed -i -e 1,8192d genlist.txt");
}
    
    // INIT
    for(int i = 0; i < DIM; i++) {
      for(int j=0; j<NLIMBS; j++) {
        outbuf[i*NLIMBS + j] = 0;
        if(j > 0) {
          outbuf[i*NLIMBS + j] = i;
        }
      }
    }

    CUDA_MEMCPY_HTD(dev_outbuf, outbuf, DIM * NLIMBS * sizeof(unsigned int));
    CUDA_MEMCPY_HTD(dev_outcnt, &outcnt, sizeof(unsigned int));
    CUDA_MEMCPY_HTD(dev_primes, primes, NPRIMES * sizeof(unsigned int));
    CUDA_MEMCPY_HTD(dev_list, enum_list, BLOCKS * TOTAL * 2 * sizeof(uint32_t));

    // all set. do it.
    printf ("Start!\n");
    add<<<BLOCKS,TOTAL>>>(dev_outbuf, dev_outcnt, dev_list, dev_primes);

    CUDA_MEMCPY_DTH(outbuf, dev_outbuf, DIM * NLIMBS * sizeof(unsigned int));

    for(unsigned int i = 0; i < DIM; i++) {
      if(outbuf[i*NLIMBS + 1] == i && outbuf[i*NLIMBS + 2] == i && outbuf[i*NLIMBS + 3] == i) continue;
      for(int j=NLIMBS-2; j >= 0; j--) {
        printf("%08x", outbuf[i*NLIMBS + j]);
      }
      printf("\n");
    }

  CUDA_FREE(dev_outbuf);
  CUDA_FREE(dev_outcnt);
  CUDA_FREE(dev_primes);

    return 0;
}


